// File Sat/FormulaCNF.java
import java.util.*;
import java.io.*;
public class FormulaCNF implements Cloneable {
    // CAMPI DATI
    private static FileIn file;   // file di input logico 
    private HashSet<Clausola> costituitaDa;
    // COSTRUTTORE
    public FormulaCNF() {
        costituitaDa = new HashSet<Clausola>();
    }
    // GESTIONE ASSOCIAZIONE
    public void inserisciLinkCostituitaDa(Clausola cla) {
        if (cla != null) costituitaDa.add(cla);
    }
    public void eliminaLinkCostituitaDa(Clausola cla) {
        if (cla != null) costituitaDa.remove(cla);
    }
    public Set<Clausola> getCostituitaDa() {
        return (HashSet<Clausola>)costituitaDa.clone();
    }
    // OPERAZIONI DELLA CLASSE
    public boolean vuota() {
        return costituitaDa.isEmpty();
    }
    public boolean contieneClausolaVuota() {
        Iterator<Clausola> itc = getCostituitaDa().iterator();
        while(itc.hasNext()) {
            Clausola elemc = itc.next();
            if (elemc.isEmpty())
                return true;
        }
        return false;
    }
    public Set<Variabile> alfabeto() {
        HashSet<Variabile> result = new HashSet();
        Iterator<Clausola> itc = getCostituitaDa().iterator();
        while(itc.hasNext()) {
            Clausola elemc = itc.next();
            Iterator<Letterale> itl = elemc.getContiene().iterator();
            while(itl.hasNext()) {
                Letterale eleml = itl.next();
                result.add(eleml.getVariabile());
            }
        }
        return result;
    }
    public FormulaCNF semplifica(Assegnazione a) {
        // ALGORITMO
        // 1. si fa una copia della FormulaCNF (e da ora in poi si 
        //    modifica solo la copia)  
        // 2. si elimina ogni clausola che non  soddisfatta da a
        // 3. si elimina da ogni clausola ogni letterale 
        //    che  reso falso da a
        // 4. si restituisce la copia modificata
        // 1.
        FormulaCNF result = (FormulaCNF)this.clone();
        // 2.
        Iterator<Clausola> itc = result.getCostituitaDa().iterator();
        while(itc.hasNext()) {
            Clausola elemc = itc.next();
            Iterator<Letterale> itl = elemc.getContiene().iterator();
            while(itl.hasNext()) {
                Letterale eleml = itl.next();
                Iterator<Letterale> ita = a.getDi().iterator();
                while(ita.hasNext()) {
                    Letterale elema = ita.next();
                    if(elema.getVariabile().equals(eleml.getVariabile()) &&
                       (elema.getSegno() == eleml.getSegno()))
                        result.eliminaLinkCostituitaDa(elemc);
                }
            }
        }
        // 3.
        itc = result.getCostituitaDa().iterator();
        while(itc.hasNext()) {
            Clausola elemc = itc.next();
            Clausola copia = (Clausola)elemc.clone();
            boolean daAccorciare = false;
            Iterator<Letterale> itl = elemc.getContiene().iterator();
            while(itl.hasNext()) {
                Letterale eleml = itl.next();
                Iterator<Letterale> ita = a.getDi().iterator();
                while(ita.hasNext()) {
                    Letterale elema = ita.next();
                    if(elema.getVariabile().equals(eleml.getVariabile()) &&
                       (elema.getSegno() != eleml.getSegno())) {
                       daAccorciare = true;
                       copia.eliminaLinkContiene(eleml);
                    }
                } // ita.hasNext()
            } // itl.hasNext()
            if (daAccorciare) {
                result.eliminaLinkCostituitaDa(elemc);
                result.inserisciLinkCostituitaDa(copia);
            }
        } // itc.hasNext()
        // 4.
        return result;
    }
    // FUNZIONI DI SERVIZIO E AUSILIARIE
    public Object clone() {
        try {
            FormulaCNF d = (FormulaCNF)super.clone(); 
            d.costituitaDa = new HashSet<Clausola>();
            Iterator<Clausola> it = getCostituitaDa().iterator();
            while(it.hasNext()) {
                Clausola elem = it.next();
                Clausola copia = (Clausola)elem.clone();
                d.inserisciLinkCostituitaDa(copia);
            }
            return d;
        } catch (CloneNotSupportedException e) {
            // non pu accadere, ma va comunque gestito
            throw new InternalError(e.toString());
        }
    }
    public String toString() {
        String result = new String("[");
        Iterator<Clausola> it = getCostituitaDa().iterator();
        while(it.hasNext()) {
            Clausola elem = it.next();
            result += elem + " ";
        }
        result += "]";
        return result;
    }
    public static FormulaCNF leggiDaFileDIMACS(String inputFile) 
        throws FileNotFoundException {
        FormulaCNF result = new FormulaCNF();
        file = new FileIn(inputFile);
        skipComments();
        String str = file.readString(); // legge il token "cnf" e lo ignora
        int nVars = file.readInt(), nClauses = file.readInt();
        LinkedList<Variabile> listaVar = new LinkedList<Variabile>();
        for (int i = 1; i <= nVars; i++) {
            Variabile v = new Variabile(String.valueOf(i));
            listaVar.add(v);
        }
        for (int i = 1; i <= nClauses; i++) {
            Clausola c = new Clausola();
            int var;
            do {
                // CICLO INTERNO: gestisce i letterali di una clausola
                // legge fino a "0", incluso
                var = file.readInt();
                if (var != 0) {
                    Letterale l;
                    if (var > 0)
                        l = new Letterale(listaVar.get(var - 1),true);
                    else
                        l = new Letterale(listaVar.get(-var - 1),false);
                    c.inserisciLinkContiene(l);
                }
            } // fine ciclo INTERNO
            while (var != 0);
            result.inserisciLinkCostituitaDa(c);
        } // fine ciclo esterno
        return result;
    }
    private static void skipComments() {
        // legge fino a "p" della "linea del problema", incluso
        char ch = file.readChar();
        if (ch == 'c') {
            file.flush();
            skipComments(); 
        }
    }
}
