// File Sat/Solutori.java
import java.util.*;

public final class Solutori {
    // CAMPI DATI
    static boolean DEBUG = false;
    // OPERAZIONI DELLA CLASSE
    public static String alfabeto2String(Set<Variabile> alfabeto) {
        String result = new String("(");
        Iterator<Variabile> it = alfabeto.iterator();
        while(it.hasNext()) {
            Variabile elem = it.next();
            result += elem.getNome() + " ";
        }
        result += ")";
        return result;
    }
    public static Assegnazione risolviConBacktracking(FormulaCNF phi) {
        // restituisce un'Assegnazione che  un modello se phi  
        // soddisfacibile, null altrimenti
        Assegnazione alpha = new Assegnazione();
        if (BT(phi,alpha))
            return alpha;
        else
            return null;
    }
    public static boolean BT(FormulaCNF phi, Assegnazione alpha) {
        // restituisce true e alpha diventa un'Assegnazione che  un modello se
        // phi  soddisfacibile, restituisce false e alpha non  significativa,
        // altrimenti
        FormulaCNF semplificazione = phi.semplifica(alpha);
        if (DEBUG) 
            System.out.println("Formula corrente dopo la semplificazione = " 
                               + semplificazione);
        if (semplificazione.vuota()) {
            if (DEBUG) 
                System.out.println("E' vuota --> SAT");
            return true;
        }
        if (semplificazione.contieneClausolaVuota()) {
            if (DEBUG) 
                System.out.println("Contiene clausola vuota --> UNSAT");
            return false;
        }
        Set alfabeto = semplificazione.alfabeto();
        if (DEBUG) 
            System.out.println("Alfabeto corrente = " + 
                               alfabeto2String(alfabeto));
        if (alfabeto.isEmpty()) {
        // non dovrebbe mai succedere che alfabeto.isEmpty()
            System.out.println("Alfabeto vuoto!");
            System.exit(-1);
        }
        Iterator<Set> its = alfabeto.iterator();
        Variabile var = (Variabile)its.next();
        Letterale p = new Letterale(var,true);
        alpha.inserisciLinkDi(p);
        if (DEBUG) {
            System.out.println("Letterale scelto per il tentativo = " + p);		
            System.out.println("Tento assegnazione POS:" + p);
            System.out.println("Assegnazione corrente = " + alpha);
        }
        if (BT(phi,alpha))
            return true;
        else {
            // elimina p da alpha
            if (!alpha.eliminaLinkDi(p))
                System.out.println("Eliminazione di " + p + " fallita!");
            p = new Letterale(var,false);	    
            alpha.inserisciLinkDi(p);
            if (DEBUG) {
                System.out.println("Tento assegnazione NEG:" + p);
                System.out.println("Assegnazione corrente = " + alpha);
            }
            boolean altroTentativo = BT(phi,alpha);
            // elimina p da alpha
            alpha.eliminaLinkDi(p);	    
            return altroTentativo;
        }
    }
    private Solutori() { };
}
